# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "Kernel", "Sum", "Product",
    "LinearKernel",
    "MyLocalGaussianKernel",
    "RationalQuadraticKernel",
    "ExpKernel",
    "LocalGaussianKernel",
    "EmptyKernel",
    "CosineKernel",
    "Matern52Kernel",
    "ExpSine2Kernel",
    "ConstantKernel",
    "ExpSquaredKernel",
    "Matern32Kernel",
    "PolynomialKernel",
    "DotProductKernel",
]

import numpy as np

from .modeling import Model, ModelSet
from .metrics import Metric, Subspace
from .kernel_interface import KernelInterface


class Kernel(ModelSet):
    """
    The abstract kernel type. Every kernel implemented in George should be
    a subclass of this object.

    """

    is_kernel = True
    kernel_type = -1

    # This function deals with weird behavior when performing arithmetic
    # operations with numpy scalars.
    def __array_wrap__(self, array, context=None):
        if context is None:
            raise TypeError("Invalid operation")
        ufunc, args, _ = context
        if ufunc.__name__ == "multiply":
            return float(args[0]) * args[1]
        elif ufunc.__name__ == "add":
            return float(args[0]) + args[1]
        raise TypeError("Invalid operation")
    __array_priority__ = np.inf

    def __getstate__(self):
        odict = self.__dict__.copy()
        odict["_kernel"] = None
        return odict

    # We must overload the ModelSet attribute getter to pass the requests
    # to the "BaseKernel"
    def __getattr__(self, name):
        if "models" in self.__dict__:
            if name in self.models:
                return self.models[name]
            if None in self.models:
                return getattr(self.models[None], name)
        raise AttributeError(name)

    @property
    def kernel(self):
        return KernelInterface(self)

    def __repr__(self):
        kernel = self.models[None]
        params = ["{0}={1}".format(k, getattr(kernel, k))
                  for k in kernel.parameter_names]
        if self.stationary:
            params += ["metric={0}".format(repr(self.metric)),
                       "block={0}".format(repr(self.block))]
        else:
            params += ["ndim={0}".format(self.ndim),
                       "axes={0}".format(repr(self.axes))]
        return "{0}({1})".format(self.__class__.__name__, ", ".join(params))

    def __add__(self, b):
        if not hasattr(b, "is_kernel"):
            return Sum(ConstantKernel(log_constant=np.log(float(b)/self.ndim),
                                      ndim=self.ndim), self)
        return Sum(self, b)

    def __radd__(self, b):
        return self.__add__(b)

    def __mul__(self, b):
        if not hasattr(b, "is_kernel"):
            log_constant = np.log(float(b)/self.ndim)
            return Product(ConstantKernel(log_constant=log_constant,
                                          ndim=self.ndim), self)
        return Product(self, b)

    def __rmul__(self, b):
        return self.__mul__(b)

    def get_value(self, x1, x2=None, diag=False):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            if diag:
                return self.kernel.value_diagonal(x1, x1)
            else:
                return self.kernel.value_symmetric(x1)
        x2 = np.ascontiguousarray(x2, dtype=np.float64)
        if diag:
            return self.kernel.value_diagonal(x1, x2)
        else:
            return self.kernel.value_general(x1, x2)

    def get_gradient(self, x1, x2=None, include_frozen=False):
        mask = (
            np.ones(self.full_size, dtype=bool)
            if include_frozen else self.unfrozen_mask
        )
        which = mask.astype(np.uint32)
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            g = self.kernel.gradient_symmetric(which, x1)
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
            g = self.kernel.gradient_general(which, x1, x2)
        return g[:, :, mask]

    def get_x1_gradient(self, x1, x2=None):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            x2 = x1
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
        return self.kernel.x1_gradient_general(x1, x2)

    def get_x2_gradient(self, x1, x2=None):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            x2 = x1
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
        return self.kernel.x2_gradient_general(x1, x2)

    def test_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        vector = self.get_parameter_vector()
        g0 = self.get_gradient(x1, x2=x2)

        for i, v in enumerate(vector):
            vector[i] = v + eps
            self.set_parameter_vector(vector)
            kp = self.get_value(x1, x2=x2)

            vector[i] = v - eps
            self.set_parameter_vector(vector)
            km = self.get_value(x1, x2=x2)

            vector[i] = v
            self.set_parameter_vector(vector)

            grad = 0.5 * (kp - km) / eps
            assert np.allclose(g0[:, :, i], grad, **kwargs), \
                "incorrect gradient for parameter '{0}' ({1})" \
                .format(self.get_parameter_names()[i], i)

    def test_x1_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        kwargs["atol"] = kwargs.get("atol", 0.5 * eps)
        g0 = self.get_x1_gradient(x1, x2=x2)
        if x2 is None:
            x2 = np.array(x1)
        for i in range(len(x1)):
            for k in range(self.ndim):
                x1[i, k] += eps
                kp = self.get_value(x1, x2=x2)

                x1[i, k] -= 2*eps
                km = self.get_value(x1, x2=x2)

                x1[i, k] += eps

                grad = 0.5 * (kp - km) / eps
                assert np.allclose(g0[i, :, k], grad[i], **kwargs)

    def test_x2_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        kwargs["atol"] = kwargs.get("atol", 0.5 * eps)
        g0 = self.get_x2_gradient(x1, x2=x2)
        if x2 is None:
            x2 = np.array(x1)
        for i in range(len(x2)):
            for k in range(self.ndim):
                x2[i, k] += eps
                kp = self.get_value(x1, x2=x2)

                x2[i, k] -= 2*eps
                km = self.get_value(x1, x2=x2)

                x2[i, k] += eps

                grad = 0.5 * (kp - km) / eps
                assert np.allclose(g0[:, i, k], grad[:, i], **kwargs)


class _operator(Kernel):
    is_kernel = False
    kernel_type = -1
    operator_type = -1

    def __init__(self, k1, k2):
        if k1.ndim != k2.ndim:
            raise ValueError("Dimension mismatch")
        self.ndim = k1.ndim
        self._dirty = True
        super(_operator, self).__init__([("k1", k1), ("k2", k2)])

    @property
    def k1(self):
        return self.models["k1"]

    @property
    def k2(self):
        return self.models["k2"]

    @property
    def dirty(self):
        return self._dirty or self.k1.dirty or self.k2.dirty

    @dirty.setter
    def dirty(self, v):
        self._dirty = v
        self.k1.dirty = False
        self.k2.dirty = False


class Sum(_operator):
    is_kernel = False
    operator_type = 0

    def __repr__(self):
        return "{0} + {1}".format(self.k1, self.k2)


class Product(_operator):
    is_kernel = False
    operator_type = 1

    def __repr__(self):
        return "{0} * {1}".format(self.k1, self.k2)


class BaseLinearKernel (Model):
    parameter_names = ("log_gamma2", )


class LinearKernel (Kernel):
    r"""
    The linear regression kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) =
            \frac{(\mathbf{x}_i \cdot \mathbf{x}_j)^P}{\gamma^2}

    :param order:
        The power :math:`P`. This parameter is a *constant*; it is not
        included in the parameter vector.

    :param log_gamma2:
        The scale factor :math:`\gamma^2`.


    """

    kernel_type = 0
    stationary = False

    def __init__(self,
                 log_gamma2=None,
                 order=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if order is None:
            raise ValueError("missing required parameter 'order'")
        self.order = order
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(log_gamma2=log_gamma2, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseLinearKernel(**kwargs)
        super(LinearKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseMyLocalGaussianKernel (Model):
    parameter_names = ("x0", "log_w", )


class MyLocalGaussianKernel (Kernel):
    r"""
    You should always document your code.

    """

    kernel_type = 1
    stationary = False

    def __init__(self,
                 x0=None,
                 log_w=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(x0=x0, log_w=log_w, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseMyLocalGaussianKernel(**kwargs)
        super(MyLocalGaussianKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseRationalQuadraticKernel (Model):
    parameter_names = ("log_alpha", )


class RationalQuadraticKernel (Kernel):
    r"""
    This is equivalent to a "scale mixture" of :class:`ExpSquaredKernel`
    kernels with different scale lengths drawn from a gamma distribution.
    See R&W for more info but here's the equation:

    .. math::
        k(r^2) = \left[1 - \frac{r^2}{2\,\alpha} \right]^\alpha

    :param log_alpha:
        The Gamma distribution parameter.


    """

    kernel_type = 2
    stationary = True

    def __init__(self,
                 log_alpha=None,
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block

        kwargs = dict(log_alpha=log_alpha, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseRationalQuadraticKernel(**kwargs)
        super(RationalQuadraticKernel, self).__init__([
            (None, base), ("metric", metric)
        ])

        # Common setup.
        self.dirty = True
    
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    

class BaseExpKernel (Model):
    parameter_names = ()


class ExpKernel (Kernel):
    r"""
    The exponential kernel is a stationary kernel where the value
    at a given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \exp \left ( -\sqrt{r^2} \right )


    """

    kernel_type = 3
    stationary = True

    def __init__(self,
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseExpKernel(**kwargs)
        super(ExpKernel, self).__init__([
            (None, base), ("metric", metric)
        ])

        # Common setup.
        self.dirty = True
    
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    

class BaseLocalGaussianKernel (Model):
    parameter_names = ("location", "log_width", )


class LocalGaussianKernel (Kernel):
    r"""
    A local Gaussian kernel.

    .. math::
        k(\mathbf{x}_i,\,\mathbf{x}_j) = \exp\left(
            -\frac{(x_i - x_0)^2 + (x_j - x_0)^2}{2\,w} \right))

    :param location:
        The location :math:`x_0` of the Gaussian.

    :param log_width:
        The (squared) width :math:`w` of the Gaussian.


    """

    kernel_type = 4
    stationary = False

    def __init__(self,
                 location=None,
                 log_width=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(location=location, log_width=log_width, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseLocalGaussianKernel(**kwargs)
        super(LocalGaussianKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseEmptyKernel (Model):
    parameter_names = ()


class EmptyKernel (Kernel):
    r"""
    This kernel is a no-op

    """

    kernel_type = 5
    stationary = False

    def __init__(self,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseEmptyKernel(**kwargs)
        super(EmptyKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseCosineKernel (Model):
    parameter_names = ("log_period", )


class CosineKernel (Kernel):
    r"""
    The simplest periodic kernel. This

    .. math::
        k(\mathbf{x}_i,\,\mathbf{x}_j) = \cos\left(
            \frac{2\,\pi\,|x_i - x_j|}{P} \right)

    where the parameter :math:`P` is the period of the oscillation. This
    kernel should probably always be multiplied be a stationary kernel
    (e.g. :class:`ExpSquaredKernel`) to allow quasi-periodic variations.

    :param log_period:
        The period of the oscillation.


    """

    kernel_type = 6
    stationary = False

    def __init__(self,
                 log_period=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(log_period=log_period, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseCosineKernel(**kwargs)
        super(CosineKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseMatern52Kernel (Model):
    parameter_names = ()


class Matern52Kernel (Kernel):
    r"""
    The Matern-5/2 kernel is stationary kernel where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \left( 1+\sqrt{5\,r^2}+ \frac{5\,r^2}{3} \right)\,
                 \exp \left (-\sqrt{5\,r^2} \right )


    """

    kernel_type = 7
    stationary = True

    def __init__(self,
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseMatern52Kernel(**kwargs)
        super(Matern52Kernel, self).__init__([
            (None, base), ("metric", metric)
        ])

        # Common setup.
        self.dirty = True
    
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    

class BaseExpSine2Kernel (Model):
    parameter_names = ("gamma", "log_period", )


class ExpSine2Kernel (Kernel):
    r"""
    The exp-sine-squared kernel is a commonly used periodic kernel. Unlike
    the :class:`CosineKernel`, this kernel never has negative covariance
    which might be useful for your problem. Here's the equation:

    .. math::
        k(\mathbf{x}_i,\,\mathbf{x}_j) =
            \exp \left( -\Gamma\,\sin^2\left[
                \frac{\pi}{P}\,\left|x_i-x_j\right|
            \right] \right)

    :param gamma:
        The scale :math:`\Gamma` of the correlations.

    :param log_period:
        The log of the period :math:`P` of the oscillation (in the same units
        as :math:`\mathbf{x}`).


    """

    kernel_type = 8
    stationary = False

    def __init__(self,
                 gamma=None,
                 log_period=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(gamma=gamma, log_period=log_period, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseExpSine2Kernel(**kwargs)
        super(ExpSine2Kernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseConstantKernel (Model):
    parameter_names = ("log_constant", )


class ConstantKernel (Kernel):
    r"""
    This kernel returns the constant

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = c

    where :math:`c` is a parameter.

    :param log_constant:
        The log of :math:`c` in the above equation.


    """

    kernel_type = 9
    stationary = False

    def __init__(self,
                 log_constant=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(log_constant=log_constant, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseConstantKernel(**kwargs)
        super(ConstantKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseExpSquaredKernel (Model):
    parameter_names = ()


class ExpSquaredKernel (Kernel):
    r"""
    The exponential-squared kernel is a stationary kernel where the value
    at a given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \exp \left ( -\frac{r^2}{2} \right )


    """

    kernel_type = 10
    stationary = True

    def __init__(self,
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseExpSquaredKernel(**kwargs)
        super(ExpSquaredKernel, self).__init__([
            (None, base), ("metric", metric)
        ])

        # Common setup.
        self.dirty = True
    
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    

class BaseMatern32Kernel (Model):
    parameter_names = ()


class Matern32Kernel (Kernel):
    r"""
    The Matern-3/2 kernel is stationary kernel where the value at a
    given radius :math:`r^2` is given by:

    .. math::

        k(r^2) = \left( 1+\sqrt{3\,r^2} \right)\,
                 \exp \left (-\sqrt{3\,r^2} \right )


    """

    kernel_type = 11
    stationary = True

    def __init__(self,
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseMatern32Kernel(**kwargs)
        super(Matern32Kernel, self).__init__([
            (None, base), ("metric", metric)
        ])

        # Common setup.
        self.dirty = True
    
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    

class BasePolynomialKernel (Model):
    parameter_names = ("log_sigma2", )


class PolynomialKernel (Kernel):
    r"""
    A polynomial kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) =
            (\mathbf{x}_i \cdot \mathbf{x}_j + \sigma^2)^P

    :param order:
        The power :math:`P`. This parameter is a *constant*; it is not
        included in the parameter vector.

    :param log_sigma2:
        The variance :math:`\sigma^2 > 0`.


    """

    kernel_type = 12
    stationary = False

    def __init__(self,
                 log_sigma2=None,
                 order=None,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        if order is None:
            raise ValueError("missing required parameter 'order'")
        self.order = order
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict(log_sigma2=log_sigma2, )
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BasePolynomialKernel(**kwargs)
        super(PolynomialKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    

class BaseDotProductKernel (Model):
    parameter_names = ()


class DotProductKernel (Kernel):
    r"""
    The dot product kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = \mathbf{x}_i \cdot \mathbf{x}_j

    with no parameters.


    """

    kernel_type = 13
    stationary = False

    def __init__(self,
                 bounds=None,
                 ndim=1,
                 axes=None):
        
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes

        kwargs = dict()
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = BaseDotProductKernel(**kwargs)
        super(DotProductKernel, self).__init__([
            (None, base), 
        ])

        # Common setup.
        self.dirty = True
    



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["HODLRSolver"]

import numpy as np

from .basic import BasicSolver
from ._hodlr import HODLRSolver as HODLRSolverInterface


class HODLRSolver(BasicSolver):

    def __init__(self, kernel, min_size=100, tol=0.1, seed=42):
        self.min_size = min_size
        self.tol = tol
        self.seed = seed
        super(HODLRSolver, self).__init__(kernel)

    def compute(self, x, yerr):
        self.solver = HODLRSolverInterface()
        self.solver.compute(self.kernel, x, yerr,
                            self.min_size, self.tol, self.seed)
        self._log_det = self.solver.log_determinant
        self.computed = self.solver.computed

    def apply_inverse(self, y, in_place=False):
        return self.solver.apply_inverse(y, in_place=in_place)

    def dot_solve(self, y):
        return self.solver.dot_solve(y)

    def apply_sqrt(self, r):
        raise NotImplementedError("apply_sqrt is not implemented for the "
                                  "HODLRSolver")

    def get_inverse(self):
        return self.solver.get_inverse()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_computed"] = False
        del state["solver"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["GP"]

import warnings
import numpy as np
from scipy.linalg import LinAlgError

from . import kernels
from .solvers import TrivialSolver, BasicSolver
from .modeling import ModelSet, ConstantModel
from .utils import multivariate_gaussian_samples, nd_sort_samples


# MAGIC: tiny epsilon to add on the diagonal of the matrices in the absence
# of observational uncertainties. Needed for computational stability.
TINY = 1.25e-12


class GP(ModelSet):
    """
    The basic Gaussian Process object.

    :param kernel:
        An instance of a subclass of :class:`kernels.Kernel`. If no kernel
        is specified then an EmptyKernel shall be used as default.

    :param fit_kernel: (optional)
        If ``True``, the parameters of the kernel will be included in all
        the relevant methods (:func:`get_parameter_vector`,
        :func:`grad_log_likelihood`, etc.). (default: ``True``)

    :param mean: (optional)
        A description of the mean function. See :py:attr:`mean` for more
        information. (default: ``0.0``)

    :param fit_mean: (optional)
        If ``True``, the parameters of the mean function will be included in
        all the relevant methods (:func:`get_parameter_vector`,
        :func:`grad_log_likelihood`, etc.). (default: ``False``)

    :param white_noise: (optional)
        A description of the logarithm of the white noise variance added to
        the diagonal of the covariance matrix. See :py:attr:`white_noise` for
        more information. (default: ``log(TINY)``)

    :param fit_white_noise: (optional)
        If ``True``, the parameters of :py:attr:`white_noise` will be included
        in all the relevant methods (:func:`get_parameter_vector`,
        :func:`grad_log_likelihood`, etc.). (default: ``False``)

    :param solver: (optional)
        The solver to use for linear algebra as documented in :ref:`solvers`.

    :param kwargs: (optional)
        Any additional arguments are passed directly to the solver's init
        function.

    """

    def __init__(
        self,
        kernel=None,
        fit_kernel=True,
        mean=None,
        fit_mean=None,
        white_noise=None,
        fit_white_noise=None,
        solver=None,
        **kwargs
    ):
        self._computed = False
        self._alpha = None
        self._y = None

        super(GP, self).__init__(
            [
                (
                    "mean",
                    ConstantModel(0.0) if mean is None else _parse_model(mean),
                ),
                (
                    "white_noise",
                    ConstantModel(np.log(TINY))
                    if white_noise is None
                    else _parse_model(white_noise),
                ),
                (
                    "kernel",
                    kernels.EmptyKernel() if kernel is None else kernel,
                ),
            ]
        )

        # Default behavior for fit_mean and fit_white_noise. If a constant is
        # provided and the fit_* parameter is not, we shouldn't try to fit
        # that component.
        try:
            float(mean)
        except TypeError:
            pass
        else:
            fit_mean = False if fit_mean is None else fit_mean

        try:
            float(white_noise)
        except TypeError:
            pass
        else:
            fit_white_noise = (
                False if fit_white_noise is None else fit_white_noise
            )

        if not fit_kernel:
            self.models["kernel"].freeze_all_parameters()
        if mean is None or (fit_mean is not None and not fit_mean):
            self.models["mean"].freeze_all_parameters()
        if white_noise is None or (
            fit_white_noise is not None and not fit_white_noise
        ):
            self.models["white_noise"].freeze_all_parameters()

        if solver is None:
            trivial = (
                kernel is None
                or kernel.kernel_type == kernels.EmptyKernel.kernel_type
            )
            solver = TrivialSolver if trivial else BasicSolver
        self.solver_type = solver
        self.solver_kwargs = kwargs
        self.solver = None

    @property
    def mean(self):
        """
        An object (following the modeling protocol) that specifies the mean
        function of the GP. You can safely set this to a scalar, a callable,
        or an instance of a class satisfying the modeling protocol. In each
        case, the mean will be evaluated (either by calling the function or
        evaluating the :func:`get_value` method) at the input coordinates and
        it should return the one-dimensional mean evaluated at these
        coordinates.

        """
        return self.models["mean"]

    def _call_mean(self, x):
        if len(x.shape) == 2 and x.shape[1] == 1:
            mu = self.mean.get_value(x[:, 0]).flatten()
        else:
            mu = self.mean.get_value(x).flatten()
        if not np.all(np.isfinite(mu)):
            raise ValueError(
                "mean function returned NaN or Inf for "
                "parameters:\n{0}".format(
                    self.mean.get_parameter_dict(include_frozen=True)
                )
            )
        return mu

    def _call_mean_gradient(self, x):
        if len(x.shape) == 2 and x.shape[1] == 1:
            mu = self.mean.get_gradient(x[:, 0])
        else:
            mu = self.mean.get_gradient(x)
        if np.any(np.isnan(mu)) or np.any(np.isinf(mu)):
            raise ValueError(
                "mean gradient function returned NaN or Inf for "
                "parameters:\n{0}".format(
                    self.mean.get_parameter_dict(include_frozen=True)
                )
            )
        return mu

    @property
    def white_noise(self):
        """
        An object (following the modeling protocol) that specifies the
        natural logarithm of the white noise variance added to the diagonal of
        the covariance matrix. You can safely set this to a scalar, a callable,
        or an instance of a class satisfying the modeling protocol. In each
        case, it will be evaluated (either by calling the function or
        evaluating the :func:`get_value` method) at the input coordinates and
        it should return the one-dimensional log-variance evaluated at these
        coordinates.

        This functionality is preferred to the ``WhiteKernel`` class provided
        by earlier versions of this module.

        """
        return self.models["white_noise"]

    def _call_white_noise(self, x):
        if len(x.shape) == 2 and x.shape[1] == 1:
            return self.white_noise.get_value(x[:, 0]).flatten()
        return self.white_noise.get_value(x).flatten()

    def _call_white_noise_gradient(self, x):
        if len(x.shape) == 2 and x.shape[1] == 1:
            return self.white_noise.get_gradient(x[:, 0])
        return self.white_noise.get_gradient(x)

    @property
    def computed(self):
        """
        Has the processes been computed since the last update of the kernel?

        """
        return (
            self._computed
            and self.solver.computed
            and (self.kernel is None or not self.kernel.dirty)
        )

    @computed.setter
    def computed(self, v):
        self._computed = v
        if v and self.kernel is not None:
            self.kernel.dirty = False

    def parse_samples(self, t):
        """
        Parse a list of samples to make sure that it has the correct
        dimensions.

        :param t: ``(nsamples,)`` or ``(nsamples, ndim)``
            The list of samples. If 1-D, this is assumed to be a list of
            one-dimensional samples otherwise, the size of the second
            dimension is assumed to be the dimension of the input space.

        Raises:
            ValueError: If the input dimension doesn't match the dimension of
                the kernel.

        """
        t = np.atleast_1d(t)
        # Deal with one-dimensional data.
        if len(t.shape) == 1:
            t = np.atleast_2d(t).T

        # Double check the dimensions against the kernel.
        if len(t.shape) != 2 or (
            self.kernel is not None and t.shape[1] != self.kernel.ndim
        ):
            raise ValueError("Dimension mismatch")

        return t

    def _check_dimensions(self, y, check_dim=True):
        n, ndim = self._x.shape
        y = np.atleast_1d(y)
        if check_dim and len(y.shape) > 1:
            raise ValueError("The predicted dimension must be 1-D")
        if len(y) != n:
            raise ValueError("Dimension mismatch")
        return y

    def _compute_alpha(self, y, cache):
        # Recalculate alpha only if y is not the same as the previous y.
        if not cache:
            r = np.ascontiguousarray(
                self._check_dimensions(y) - self._call_mean(self._x),
                dtype=np.float64,
            )
            return self.solver.apply_inverse(r, in_place=True).flatten()
        if self._alpha is None or not np.array_equiv(y, self._y):
            self._y = y
            r = np.ascontiguousarray(
                self._check_dimensions(y) - self._call_mean(self._x),
                dtype=np.float64,
            )
            self._alpha = self.solver.apply_inverse(r, in_place=True).flatten()
        return self._alpha

    def apply_inverse(self, y):
        """
        Self-consistently apply the inverse of the computed kernel matrix to
        some vector or matrix of samples. This method subtracts the mean,
        sorts the samples, then returns the samples in the correct (unsorted)
        order.

        :param y: ``(nsamples, )`` or ``(nsamples, K)``
            The vector (or matrix) of sample values.

        """
        self.recompute(quiet=False)
        r = np.array(y, dtype=np.float64, order="F")
        r = self._check_dimensions(r, check_dim=False)

        # Broadcast the mean function
        m = [slice(None)] + [np.newaxis for _ in range(len(r.shape) - 1)]
        r -= self._call_mean(self._x)[tuple(m)]

        # Do the solve
        if len(r.shape) == 1:
            b = self.solver.apply_inverse(r, in_place=True).flatten()
        else:
            b = self.solver.apply_inverse(r, in_place=True)
        return b

    def compute(self, x, yerr=0.0, **kwargs):
        """
        Pre-compute the covariance matrix and factorize it for a set of times
        and uncertainties.

        :param x: ``(nsamples,)`` or ``(nsamples, ndim)``
            The independent coordinates of the data points.

        :param yerr: (optional) ``(nsamples,)`` or scalar
            The Gaussian uncertainties on the data points at coordinates
            ``x``. These values will be added in quadrature to the diagonal of
            the covariance matrix.

        """
        # Parse the input coordinates and ensure the right memory layout.
        self._x = self.parse_samples(x)
        self._x = np.ascontiguousarray(self._x, dtype=np.float64)
        try:
            self._yerr2 = float(yerr) ** 2 * np.ones(len(x))
        except TypeError:
            self._yerr2 = self._check_dimensions(yerr) ** 2
        self._yerr2 = np.ascontiguousarray(self._yerr2, dtype=np.float64)

        # Set up and pre-compute the solver.
        self.solver = self.solver_type(self.kernel, **(self.solver_kwargs))

        # Include the white noise term.
        yerr = np.sqrt(self._yerr2 + np.exp(self._call_white_noise(self._x)))
        self.solver.compute(self._x, yerr, **kwargs)

        self._const = -0.5 * (
            len(self._x) * np.log(2 * np.pi) + self.solver.log_determinant
        )
        self.computed = True
        self._alpha = None

    def recompute(self, quiet=False, **kwargs):
        """
        Re-compute a previously computed model. You might want to do this if
        the kernel parameters change and the kernel is labeled as ``dirty``.

        :param quiet: (optional)
            If ``True``, return false when the computation fails. Otherwise,
            throw an error if something goes wrong. (default: ``False``)

        """
        if not self.computed:
            if not (hasattr(self, "_x") and hasattr(self, "_yerr2")):
                raise RuntimeError("You need to compute the model first")
            try:
                # Update the model making sure that we store the original
                # ordering of the points.
                self.compute(self._x, np.sqrt(self._yerr2), **kwargs)
            except (ValueError, LinAlgError):
                if quiet:
                    return False
                raise
        return True

    def lnlikelihood(self, y, quiet=False):
        warnings.warn(
            "'lnlikelihood' is deprecated. Use 'log_likelihood'",
            DeprecationWarning,
        )
        return self.log_likelihood(y, quiet=quiet)

    def log_likelihood(self, y, quiet=False):
        """
        Compute the logarithm of the marginalized likelihood of a set of
        observations under the Gaussian process model. You must call
        :func:`GP.compute` before this function.

        :param y: ``(nsamples, )``
            The observations at the coordinates provided in the ``compute``
            step.

        :param quiet:
            If ``True`` return negative infinity instead of raising an
            exception when there is an invalid kernel or linear algebra
            failure. (default: ``False``)

        """
        if not self.recompute(quiet=quiet):
            return -np.inf
        try:
            mu = self._call_mean(self._x)
        except ValueError:
            if quiet:
                return -np.inf
            raise
        r = np.ascontiguousarray(
            self._check_dimensions(y) - mu, dtype=np.float64
        )
        ll = self._const - 0.5 * self.solver.dot_solve(r)
        return ll if np.isfinite(ll) else -np.inf

    def grad_lnlikelihood(self, y, quiet=False):
        warnings.warn(
            "'grad_lnlikelihood' is deprecated. " "Use 'grad_log_likelihood'",
            DeprecationWarning,
        )
        return self.grad_log_likelihood(y, quiet=quiet)

    def grad_log_likelihood(self, y, quiet=False):
        """
        Compute the gradient of :func:`GP.log_likelihood` as a function of the
        parameters returned by :func:`GP.get_parameter_vector`. You must call
        :func:`GP.compute` before this function.

        :param y: ``(nsamples,)``
            The list of observations at coordinates ``x`` provided to the
            :func:`compute` function.

        :param quiet:
            If ``True`` return a gradient of zero instead of raising an
            exception when there is an invalid kernel or linear algebra
            failure. (default: ``False``)

        """
        # Make sure that the model is computed and try to recompute it if it's
        # dirty.
        if not self.recompute(quiet=quiet):
            return np.zeros(len(self), dtype=np.float64)

        # Pre-compute some factors.
        try:
            alpha = self._compute_alpha(y, False)
        except ValueError:
            if quiet:
                return np.zeros(len(self), dtype=np.float64)
            raise

        if len(self.white_noise) or len(self.kernel):
            K_inv = self.solver.get_inverse()
            A = np.einsum("i,j", alpha, alpha) - K_inv

        # Compute each component of the gradient.
        grad = np.empty(len(self))
        n = 0

        l = len(self.mean)
        if l:
            try:
                mu = self._call_mean_gradient(self._x)
            except ValueError:
                if quiet:
                    return np.zeros(len(self), dtype=np.float64)
                raise
            grad[n : n + l] = np.dot(mu, alpha)
            n += l

        l = len(self.white_noise)
        if l:
            wn = self._call_white_noise(self._x)
            wng = self._call_white_noise_gradient(self._x)
            grad[n : n + l] = 0.5 * np.sum(
                (np.exp(wn) * np.diag(A))[None, :] * wng, axis=1
            )
            n += l

        l = len(self.kernel)
        if l:
            Kg = self.kernel.get_gradient(self._x)
            grad[n : n + l] = 0.5 * np.einsum("ijk,ij", Kg, A)

        return grad

    def nll(self, vector, y, quiet=True):
        self.set_parameter_vector(vector)
        if not np.isfinite(self.log_prior()):
            return np.inf
        return -self.log_likelihood(y, quiet=quiet)

    def grad_nll(self, vector, y, quiet=True):
        self.set_parameter_vector(vector)
        if not np.isfinite(self.log_prior()):
            return np.zeros(len(vector))
        return -self.grad_log_likelihood(y, quiet=quiet)

    def predict(
        self, y, t, return_cov=True, return_var=False, cache=True, kernel=None
    ):
        """
        Compute the conditional predictive distribution of the model. You must
        call :func:`GP.compute` before this function.

        :param y: ``(nsamples,)``
            The observations to condition the model on.

        :param t: ``(ntest,)`` or ``(ntest, ndim)``
            The coordinates where the predictive distribution should be
            computed.

        :param return_cov: (optional)
            If ``True``, the full covariance matrix is computed and returned.
            Otherwise, only the mean prediction is computed. (default:
            ``True``)

        :param return_var: (optional)
            If ``True``, only return the diagonal of the predictive covariance;
            this will be faster to compute than the full covariance matrix.
            This overrides ``return_cov`` so, if both are set to ``True``,
            only the diagonal is computed. (default: ``False``)

        :param cache: (optional)
            If ``True`` the value of alpha will be cached to speed up repeated
            predictions.

        :param kernel: (optional)
            If provided, this kernel will be used to calculate the cross terms.
            This can be used to separate the predictions from different
            kernels.

        Returns ``mu``, ``(mu, cov)``, or ``(mu, var)`` depending on the values
        of ``return_cov`` and ``return_var``. These output values are:

        * **mu** ``(ntest,)``: mean of the predictive distribution,
        * **cov** ``(ntest, ntest)``: the predictive covariance matrix, and
        * **var** ``(ntest,)``: the diagonal elements of ``cov``.

        """
        self.recompute()
        alpha = self._compute_alpha(y, cache)
        xs = self.parse_samples(t)

        if kernel is None:
            kernel = self.kernel

        # Compute the predictive mean.
        Kxs = kernel.get_value(xs, self._x)
        mu = np.dot(Kxs, alpha) + self._call_mean(xs)
        if not (return_var or return_cov):
            return mu

        KinvKxs = self.solver.apply_inverse(Kxs.T)
        if return_var:
            var = kernel.get_value(xs, diag=True)
            var -= np.sum(Kxs.T * KinvKxs, axis=0)
            return mu, var

        cov = kernel.get_value(xs)
        cov -= np.dot(Kxs, KinvKxs)
        return mu, cov

    def sample_conditional(self, y, t, size=1):
        """
        Draw samples from the predictive conditional distribution. You must
        call :func:`GP.compute` before this function.

        :param y: ``(nsamples, )``
            The observations to condition the model on.

        :param t: ``(ntest, )`` or ``(ntest, ndim)``
            The coordinates where the predictive distribution should be
            computed.

        :param size: (optional)
            The number of samples to draw. (default: ``1``)

        Returns **samples** ``(size, ntest)``, a list of predictions at
        coordinates given by ``t``.

        """
        mu, cov = self.predict(y, t)
        return multivariate_gaussian_samples(cov, size, mean=mu)

    def sample(self, t=None, size=1):
        """
        Draw samples from the prior distribution.

        :param t: ``(ntest, )`` or ``(ntest, ndim)`` (optional)
            The coordinates where the model should be sampled. If no
            coordinates are given, the precomputed coordinates and
            factorization are used.

        :param size: (optional)
            The number of samples to draw. (default: ``1``)

        Returns **samples** ``(size, ntest)``, a list of predictions at
        coordinates given by ``t``. If ``size == 1``, the result is a single
        sample with shape ``(ntest,)``.

        """
        if t is None:
            self.recompute()
            n, _ = self._x.shape

            # Generate samples using the precomputed factorization.
            results = self.solver.apply_sqrt(np.random.randn(size, n))
            results += self._call_mean(self._x)
            return results[0] if size == 1 else results

        x = self.parse_samples(t)
        cov = self.get_matrix(x)
        cov[np.diag_indices_from(cov)] += TINY
        return multivariate_gaussian_samples(
            cov, size, mean=self._call_mean(x)
        )

    def get_matrix(self, x1, x2=None):
        """
        Get the covariance matrix at a given set or two of independent
        coordinates.

        :param x1: ``(nsamples,)`` or ``(nsamples, ndim)``
            A list of samples.

        :param x2: ``(nsamples,)`` or ``(nsamples, ndim)`` (optional)
            A second list of samples. If this is given, the cross covariance
            matrix is computed. Otherwise, the auto-covariance is evaluated.

        """
        x1 = self.parse_samples(x1)
        if x2 is None:
            return self.kernel.get_value(x1)
        x2 = self.parse_samples(x2)
        return self.kernel.get_value(x1, x2)

    def get_value(self, *args, **kwargs):
        """
        A synonym for :func:`GP.log_likelihood` provided for consistency with
        the modeling protocol.

        """
        return self.log_likelihood(*args, **kwargs)

    def get_gradient(self, *args, **kwargs):
        """
        A synonym for :func:`GP.grad_log_likelihood` provided for consistency
        with the modeling protocol.

        """
        return self.grad_log_likelihood(*args, **kwargs)


def _parse_model(model):
    try:
        val = float(model)
    except TypeError:
        return model
    return ConstantModel(float(val))



# -*- coding: utf-8 -*-

from __future__ import division, print_function
import numpy as np
from itertools import chain
from collections import OrderedDict

__all__ = ["Model", "ModelSet", "ConstantModel", "CallableModel"]


class Model(object):
    """
    An abstract class implementing the skeleton of the modeling protocol

    Initial parameter values can either be provided as arguments in the same
    order as ``parameter_names`` or by name as keyword arguments.

    A minimal subclass of this would have the form:

    .. code-block:: python

        class CustomModel(Model):
            parameter_names = ("parameter_1", "parameter_2")

            def get_value(self, x):
                return self.parameter_1 + self.parameter_2 + x

        model = CustomModel(parameter_1=1.0, parameter_2=2.0)
        # or...
        model = CustomModel(1.0, 2.0)

    Args:
        bounds (Optional[list or dict]): Bounds can be given for each
            parameter setting their minimum and maximum allowed values.
            This parameter can either be a ``list`` (with length
            ``full_size``) or a ``dict`` with named parameters. Any parameters
            that are omitted from the ``dict`` will be assumed to have no
            bounds. These bounds can be retrieved later using the
            :func:`celerite.Model.get_parameter_bounds` method and, by
            default, they are used in the :func:`celerite.Model.log_prior`
            method.

    """

    parameter_names = tuple()

    def __init__(self, *args, **kwargs):
        self.unfrozen_mask = np.ones(self.full_size, dtype=bool)
        self.dirty = True

        # Deal with bounds
        self.parameter_bounds = []
        bounds = kwargs.pop("bounds", dict())
        try:
            # Try to treat 'bounds' as a dictionary
            for name in self.parameter_names:
                self.parameter_bounds.append(bounds.get(name, (None, None)))
        except AttributeError:
            # 'bounds' isn't a dictionary - it had better be a list
            self.parameter_bounds = list(bounds)
        if len(self.parameter_bounds) != self.full_size:
            raise ValueError("the number of bounds must equal the number of "
                             "parameters")
        if any(len(b) != 2 for b in self.parameter_bounds):
            raise ValueError("the bounds for each parameter must have the "
                             "format: '(min, max)'")

        # Parameter values can be specified as arguments or keywords
        if len(args):
            if len(args) != self.full_size:
                raise ValueError("expected {0} arguments but got {1}"
                                 .format(self.full_size, len(args)))
            if len(kwargs):
                raise ValueError("parameters must be fully specified by "
                                 "arguments or keyword arguments, not both")
            self.parameter_vector = args

        else:
            # Loop over the kwargs and set the parameter values
            params = []
            for k in self.parameter_names:
                v = kwargs.pop(k, None)
                if v is None:
                    raise ValueError("missing parameter '{0}'".format(k))
                params.append(v)
            self.parameter_vector = params

            if len(kwargs):
                raise ValueError("unrecognized parameter(s) '{0}'"
                                 .format(list(kwargs.keys())))

        # Check the initial prior value
        quiet = kwargs.get("quiet", False)
        if not quiet and not np.isfinite(self.log_prior()):
            raise ValueError("non-finite log prior value")

    def get_value(self, *args, **kwargs):
        """
        Compute the "value" of the model for the current parameters

        This method should be overloaded by subclasses to implement the actual
        functionality of the model.

        """
        raise NotImplementedError("overloaded by subclasses")

    def compute_gradient(self, *args, **kwargs):
        """
        Compute the "gradient" of the model for the current parameters

        The default implementation computes the gradients numerically using
        a first order forward scheme. For better performance, this method
        should be overloaded by subclasses. The output of this function
        should be an array where the first dimension is ``full_size``.

        """
        _EPS = 1.254e-5
        vector = self.get_parameter_vector()
        value0 = self.get_value(*args, **kwargs)
        grad = np.empty([len(vector)] + list(value0.shape), dtype=np.float64)
        for i, v in enumerate(vector):
            vector[i] = v + _EPS
            self.set_vector(vector)
            value = self.get_value(*args, **kwargs)
            vector[i] = v
            self.set_vector(vector)
            grad[i] = (value - value0) / _EPS
        return grad

    def get_gradient(self, *args, **kwargs):
        include_frozen = kwargs.pop("include_frozen", False)
        g = self.compute_gradient(*args, **kwargs)
        if include_frozen:
            return g
        return g[self.unfrozen_mask]

    def __len__(self):
        return self.vector_size

    def _get_name(self, name_or_index):
        try:
            int(name_or_index)
        except (TypeError, ValueError):
            return name_or_index
        return self.get_parameter_names()[int(name_or_index)]

    def __getitem__(self, name_or_index):
        return self.get_parameter(self._get_name(name_or_index))

    def __setitem__(self, name_or_index, value):
        return self.set_parameter(self._get_name(name_or_index), value)

    @property
    def full_size(self):
        """The total number of parameters (including frozen parameters)"""
        return len(self.parameter_names)

    @property
    def vector_size(self):
        """The number of active (or unfrozen) parameters"""
        return self.unfrozen_mask.sum()

    @property
    def parameter_vector(self):
        """An array of all parameters (including frozen parameters)"""
        return np.array([getattr(self, k) for k in self.parameter_names])

    @parameter_vector.setter
    def parameter_vector(self, v):
        if len(v) != self.full_size:
            raise ValueError("dimension mismatch")
        for k, val in zip(self.parameter_names, v):
            setattr(self, k, float(val))
        self.dirty = True

    def get_parameter_dict(self, include_frozen=False):
        """
        Get an ordered dictionary of the parameters

        Args:
            include_frozen (Optional[bool]): Should the frozen parameters be
                included in the returned value? (default: ``False``)

        """
        return OrderedDict(zip(
            self.get_parameter_names(include_frozen=include_frozen),
            self.get_parameter_vector(include_frozen=include_frozen),
        ))

    def get_parameter_names(self, include_frozen=False):
        """
        Get a list of the parameter names

        Args:
            include_frozen (Optional[bool]): Should the frozen parameters be
                included in the returned value? (default: ``False``)

        """
        if include_frozen:
            return self.parameter_names
        return tuple(p
                     for p, f in zip(self.parameter_names, self.unfrozen_mask)
                     if f)

    def get_parameter_bounds(self, include_frozen=False):
        """
        Get a list of the parameter bounds

        Args:
            include_frozen (Optional[bool]): Should the frozen parameters be
                included in the returned value? (default: ``False``)

        """
        if include_frozen:
            return self.parameter_bounds
        return list(p
                    for p, f in zip(self.parameter_bounds, self.unfrozen_mask)
                    if f)

    def get_parameter_vector(self, include_frozen=False):
        """
        Get an array of the parameter values in the correct order

        Args:
            include_frozen (Optional[bool]): Should the frozen parameters be
                included in the returned value? (default: ``False``)

        """
        if include_frozen:
            return self.parameter_vector
        return self.parameter_vector[self.unfrozen_mask]

    def set_parameter_vector(self, vector, include_frozen=False):
        """
        Set the parameter values to the given vector

        Args:
            vector (array[vector_size] or array[full_size]): The target
                parameter vector. This must be in the same order as
                ``parameter_names`` and it should only include frozen
                parameters if ``include_frozen`` is ``True``.
            include_frozen (Optional[bool]): Should the frozen parameters be
                included in the returned value? (default: ``False``)

        """
        v = self.parameter_vector
        if include_frozen:
            v[:] = vector
        else:
            v[self.unfrozen_mask] = vector
        self.parameter_vector = v
        self.dirty = True

    def check_parameter_vector(self, vector):
        # Save the original parameter vector and dirtiness
        vector0 = np.array(self.get_parameter_vector())
        dirty0 = self.dirty

        # Update the vector and compute the prior
        self.set_parameter_vector(vector)
        lp = self.log_prior()

        # Reset the parameter vector
        self.set_parameter_vector(vector0)
        self.dirty = dirty0
        return np.isfinite(lp)

    def freeze_parameter(self, name):
        """
        Freeze a parameter by name

        Args:
            name: The name of the parameter

        """
        i = self.get_parameter_names(include_frozen=True).index(name)
        self.unfrozen_mask[i] = False

    def thaw_parameter(self, name):
        """
        Thaw a parameter by name

        Args:
            name: The name of the parameter

        """
        i = self.get_parameter_names(include_frozen=True).index(name)
        self.unfrozen_mask[i] = True

    def freeze_all_parameters(self):
        """Freeze all parameters of the model"""
        self.unfrozen_mask[:] = False

    def thaw_all_parameters(self):
        """Thaw all parameters of the model"""
        self.unfrozen_mask[:] = True

    def get_parameter(self, name):
        """
        Get a parameter value by name

        Args:
            name: The name of the parameter

        """
        i = self.get_parameter_names(include_frozen=True).index(name)
        return self.get_parameter_vector(include_frozen=True)[i]

    def set_parameter(self, name, value):
        """
        Set a parameter value by name

        Args:
            name: The name of the parameter
            value (float): The new value for the parameter

        """
        i = self.get_parameter_names(include_frozen=True).index(name)
        v = self.get_parameter_vector(include_frozen=True)
        v[i] = value
        self.set_parameter_vector(v, include_frozen=True)

    def log_prior(self):
        """Compute the log prior probability of the current parameters"""
        for p, b in zip(self.parameter_vector, self.parameter_bounds):
            if b[0] is not None and p < b[0]:
                return -np.inf
            if b[1] is not None and p > b[1]:
                return -np.inf
        return 0.0

    @staticmethod
    def parameter_sort(f):
        def func(self, *args, **kwargs):
            values = f(self, *args, **kwargs)
            names = self.get_parameter_names(include_frozen=True)
            ret = [values[k] for k in names]
            # Horrible hack to only return numpy array if that's what was
            # given by the wrapped function.
            if len(ret) and type(ret[0]).__module__ == np.__name__:
                return np.vstack(ret)
            return ret
        return func


class ModelSet(Model):
    """
    An abstract wrapper for combining named :class:`Model` objects

    The parameter names of a composite model are prepended by the submodel
    name. For example:

    .. code-block:: python

        model = ModelSet([
            ("model1", Model(par1=1.0, par2=2.0)),
            ("model2", Model(par3=3.0, par4=4.0)),
        ])
        print(model.get_parameter_names())

    will print

    .. code-block:: python

        ["model1:par1", "model1:par2", "model2:par3", "model2:par4"]

    Args:
        models: This should be a list of the form: ``[("model1", Model(...)),
            ("model2", Model(...)), ...]``.

    """

    def __init__(self, models):
        self.models = OrderedDict()
        for name, model in models:
            self.models[name] = model

    def __getattr__(self, name):
        if "models" in self.__dict__ and name in self.models:
            return self.models[name]
        raise AttributeError(name)

    @property
    def dirty(self):
        return any(m.dirty for m in self.models.values())

    @dirty.setter
    def dirty(self, value):
        for m in self.models.values():
            m.dirty = value

    @property
    def full_size(self):
        return sum(m.full_size for m in self.models.values())

    @property
    def vector_size(self):
        return sum(m.vector_size for m in self.models.values())

    @property
    def unfrozen_mask(self):
        return np.concatenate([
            m.unfrozen_mask for m in self.models.values()
        ])

    @property
    def parameter_vector(self):
        return np.concatenate([
            m.parameter_vector for m in self.models.values()
        ])

    @parameter_vector.setter
    def parameter_vector(self, v):
        i = 0
        for m in self.models.values():
            l = m.full_size
            m.parameter_vector = v[i:i+l]
            i += l

    @property
    def parameter_names(self):
        return tuple(chain(*(
            map("{0}".format, m.parameter_names) if name is None else
            map("{0}:{{0}}".format(name).format, m.parameter_names)
            for name, m in self.models.items()
        )))

    @property
    def parameter_bounds(self):
        return list(chain(*(
            m.parameter_bounds for m in self.models.values()
        )))

    def _apply_to_parameter(self, func, name, *args):
        comp = name.split(":")
        model_name = comp[0]
        if model_name not in self.models:
            if None in self.models:
                model_name = None
                comp = [None] + comp
            else:
                raise ValueError("unrecognized parameter '{0}'".format(name))
        return getattr(self.models[model_name], func)(":".join(comp[1:]),
                                                      *args)

    def freeze_parameter(self, name):
        self._apply_to_parameter("freeze_parameter", name)

    def thaw_parameter(self, name):
        self._apply_to_parameter("thaw_parameter", name)

    def freeze_all_parameters(self):
        for model in self.models.values():
            model.freeze_all_parameters()

    def thaw_all_parameters(self):
        for model in self.models.values():
            model.thaw_all_parameters()

    def get_parameter(self, name):
        return self._apply_to_parameter("get_parameter", name)

    def set_parameter(self, name, value):
        self.dirty = True
        return self._apply_to_parameter("set_parameter", name, value)

    def log_prior(self):
        lp = 0.0
        for model in self.models.values():
            lp += model.log_prior()
            if not np.isfinite(lp):
                return -np.inf
        return lp


class ConstantModel(Model):
    """
    A simple concrete model with a single parameter ``value``

    Args:
        value (float): The value of the model.

    """

    parameter_names = ("value", )

    def get_value(self, x):
        return self.value + np.zeros(len(x))

    def compute_gradient(self, x):
        return np.ones((1, len(x)))


class CallableModel(Model):

    def __init__(self, function, gradient=None):
        self.function = function
        self.gradient = gradient
        super(CallableModel, self).__init__()

    def get_value(self, x):
        return self.function(x)

    def compute_gradient(self, x):
        if self.gradient is not None:
            return self.gradient(x)
        return super(CallableModel, self).compute_gradient(x)
